
/* Chrysalide - Outil d'analyse de fichiers binaires
 * move.c - mémorisation des déplacements au sein de code
 *
 * Copyright (C) 2016 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "move.h"


#include <assert.h>
#include <malloc.h>
#include <stdio.h>
#include <sys/socket.h>


#include <i18n.h>


#include "../collection-int.h"
#include "../item-int.h"
#include "../../../common/io.h"
#include "../../../gui/editem.h"
#include "../../../gtkext/gtkviewpanel.h"





/* --------------------- ELABORATION D'UN ELEMENT DE COLLECTION --------------------- */


/* Déplacement d'un emplacement à un autre dans du code (instance) */
struct _GDbMove
{
    GDbItem parent;                         /* A laisser en premier        */

    vmpa2t src;                             /* Emplacement de départ       */
    vmpa2t dest;                            /* Emplacement d'arrivée       */

};

/* Déplacement d'un emplacement à un autre dans du code (classe) */
struct _GDbMoveClass
{
    GDbItemClass parent;                    /* A laisser en premier        */

};



/* Initialise la classe des déplacements au sein de codes. */
static void g_db_move_class_init(GDbMoveClass *);

/* Initialise une bascule d'affichage pour opérande numérique. */
static void g_db_move_init(GDbMove *);

/* Supprime toutes les références externes. */
static void g_db_move_dispose(GDbMove *);

/* Procède à la libération totale de la mémoire. */
static void g_db_move_finalize(GDbMove *);

/* Effectue la comparaison entre deux changements de position. */
static gint g_db_move_cmp(GDbMove *, GDbMove *, bool);

/* Importe la définition d'un déplacement depuis un flux réseau. */
static bool g_db_move_recv_from_fd(GDbMove *, int, int);

/* Exporte la définition d'un déplacement dans un flux réseau. */
static bool g_db_move_send_to_fd(const GDbMove *, int, int);

/* Construit la description humaine d'un déplacement. */
static void g_db_move_build_label(GDbMove *);

/* Exécute un déplacement dans une zone de code. */
static bool g_db_move_run(const GDbMove *, const vmpa2t *);

/* Applique un déplacement au sein d'une zone de code. */
static bool g_db_move_apply(GDbMove *, GLoadedBinary *);

/* Annule un déplacement au sein d'une zone de code. */
static bool g_db_move_cancel(GDbMove *, GLoadedBinary *);

/* Constitue les champs destinés à une insertion / modification. */
static bool g_db_move_prepare_db_statement(const GDbMove *, bound_value **, size_t *);

/* Charge les valeurs utiles pour un déplacement dans du code. */
static bool g_db_move_load(GDbMove *, const bound_value *, size_t);




/* ---------------------- DEFINITION DE LA COLLECTION ASSOCIEE ---------------------- */


/* Collection dédiée aux basculements d'affichage (instance) */
struct _GMoveCollection
{
    GDbCollection parent;                   /* A laisser en premier        */

};

/* Collection dédiée aux basculements d'affichage (classe) */
struct _GMoveCollectionClass
{
    GDbCollectionClass parent;              /* A laisser en premier        */

};


/* Initialise la classe des collections de déplacements. */
static void g_move_collection_class_init(GMoveCollectionClass *);

/* Initialise une collection de déplacements. */
static void g_move_collection_init(GMoveCollection *);

/* Supprime toutes les références externes. */
static void g_move_collection_dispose(GMoveCollection *);

/* Procède à la libération totale de la mémoire. */
static void g_move_collection_finalize(GMoveCollection *);

/* Crée la table des basculements dans une base de données. */
static bool g_move_collection_create_db_table(const GMoveCollection *, sqlite3 *);

/* Décrit les colonnes utiles à un chargement de données. */
static bool g_move_collection_setup_load(GMoveCollection *, bound_value **, size_t *);

/* Détermine si un élément est déjà présent ou non. */
static GDbItem *g_move_collection_has_key(GMoveCollection *, va_list);



/* ---------------------------------------------------------------------------------- */
/*                       ELABORATION D'UN ELEMENT DE COLLECTION                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour un déplacement au sein de code. */
G_DEFINE_TYPE(GDbMove, g_db_move, G_TYPE_DB_ITEM);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des déplacements au sein de codes.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_db_move_class_init(GDbMoveClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GDbItemClass *item;                     /* Encore une autre vision...  */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_db_move_dispose;
    object->finalize = (GObjectFinalizeFunc)g_db_move_finalize;

    item = G_DB_ITEM_CLASS(klass);

    item->feature = DBF_MOVES;

    item->cmp = (cmp_db_item_fc)g_db_move_cmp;

    item->recv = (recv_db_item_fc)g_db_move_recv_from_fd;
    item->send = (send_db_item_fc)g_db_move_send_to_fd;

    item->build_label = (build_item_label_fc)g_db_move_build_label;
    item->apply = (run_item_fc)g_db_move_apply;
    item->cancel = (run_item_fc)g_db_move_cancel;

    item->prepare_stmt = (prepare_db_statement)g_db_move_prepare_db_statement;
    item->load = (load_db_item_fc)g_db_move_load;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise un déplacement au sein de code.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_db_move_init(GDbMove *move)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_db_move_dispose(GDbMove *move)
{
    G_OBJECT_CLASS(g_db_move_parent_class)->dispose(G_OBJECT(move));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_db_move_finalize(GDbMove *move)
{
    G_OBJECT_CLASS(g_db_move_parent_class)->finalize(G_OBJECT(move));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : src  = emplacement de départ d'un mouvement.                 *
*                dest = emplacement d'arrivée du même mouvement.              *
*                                                                             *
*  Description : Organise un déplacement d'une position à une autre.          *
*                                                                             *
*  Retour      : Déplacement mis en place ou NULL en cas d'erreur.            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDbMove *g_db_move_new(const vmpa2t *src, const vmpa2t *dest)
{
    GDbMove *result;                        /* Instance à retourner        */

    result = g_object_new(G_TYPE_DB_MOVE, NULL);

    copy_vmpa(&result->src, src);
    copy_vmpa(&result->dest, dest);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : a    = premier élément à analyser.                           *
*                b    = second élément à analyser.                            *
*                with = précise les horodatages à prendre en compte.          *
*                                                                             *
*  Description : Effectue la comparaison entre deux changements de position.  *
*                                                                             *
*  Retour      : Bilan de la comparaison : -1, 0 ou 1.                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static gint g_db_move_cmp(GDbMove *a, GDbMove *b, bool with)
{
    gint result;                            /* Bilan de la comparaison     */

    result = cmp_vmpa(&a->src, &b->src);

    if (result == 0)
        result = cmp_vmpa(&a->dest, &b->dest);

    if (result == 0)
        result = G_DB_ITEM_CLASS(g_db_move_parent_class)->cmp(G_DB_ITEM(a), G_DB_ITEM(b), with);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move  = bascule d'affichage aux infos à charger. [OUT]       *
*                fd    = flux ouvert en lecture pour l'importation.           *
*                flags = éventuelles options d'envoi supplémentaires.         *
*                                                                             *
*  Description : Importe la définition d'un déplacement depuis un flux réseau.*
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_db_move_recv_from_fd(GDbMove *move, int fd, int flags)
{
    bool status;                            /* Bilan d'opération initiale  */

    status = G_DB_ITEM_CLASS(g_db_move_parent_class)->recv(G_DB_ITEM(move), fd, flags);
    if (!status) return false;

    if (!recv_vmpa(&move->src, fd, flags))
        return false;

    if (!recv_vmpa(&move->dest, fd, flags))
        return false;

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move  = bascule d'affichage aux infos à sauvegarder.         *
*                fd    = flux ouvert en écriture pour l'exportation.          *
*                flags = éventuelles options d'envoi supplémentaires.         *
*                                                                             *
*  Description : Exporte la définition d'un déplacement dans un flux réseau.  *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_db_move_send_to_fd(const GDbMove *move, int fd, int flags)
{
    bool status;                            /* Bilan d'opération initiale  */

    status = G_DB_ITEM_CLASS(g_db_move_parent_class)->send(G_DB_ITEM(move), fd, MSG_MORE | flags);
    if (!status) return false;

    if (!send_vmpa(&move->src, fd, MSG_MORE | flags))
        return false;

    if (!send_vmpa(&move->dest, fd, flags))
        return false;

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move = informations de déplacement à manipuler.              *
*                                                                             *
*  Description : Construit la description humaine d'un déplacement.           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_db_move_build_label(GDbMove *move)
{
    VMPA_BUFFER(src_loc);                   /* Indication de position #1   */
    VMPA_BUFFER(dest_loc);                  /* Indication de position #2   */

    if (has_virt_addr(&move->src))
        vmpa2_virt_to_string(&move->src, MDS_UNDEFINED, src_loc, NULL);
    else
        vmpa2_phys_to_string(&move->src, MDS_UNDEFINED, src_loc, NULL);

    if (has_virt_addr(&move->dest))
        vmpa2_virt_to_string(&move->dest, MDS_UNDEFINED, dest_loc, NULL);
    else
        vmpa2_phys_to_string(&move->dest, MDS_UNDEFINED, dest_loc, NULL);

    asprintf(&G_DB_ITEM(move)->label, _("Move from %s to %s"), src_loc, dest_loc);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move = informations de déplacement à consulter.              *
*                addr = localisation à placer bien en évidence.               *
*                                                                             *
*  Description : Exécute un déplacement dans une zone de code.                *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_db_move_run(const GDbMove *move, const vmpa2t *addr)
{
    GtkViewPanel *vpanel;                   /* Afficheur effectif de code  */

    typedef struct _move_params
    {
        GtkViewPanel *vpanel;               /* Afficheur effectif de code  */
        vmpa2t addr;                        /* Adresse de déplacement      */

    } move_params;

    move_params *params;                    /* Paramètres à transmettre    */

    /* Procédure de nettoyage */

    void delete_move_params(move_params *p)
    {
        g_object_unref(G_OBJECT(p->vpanel));

        free(p);

    }

    /* Procédure d'application différée */

    gboolean do_move_in_main_loop(move_params *p)
    {
        gtk_view_panel_scroll_to_address(p->vpanel, &p->addr, SPT_CENTER);

        return G_SOURCE_REMOVE;

    }

    /* Lancement de l'opération */

    vpanel = g_editor_item_get_current_view(NULL);

    if (vpanel != NULL)
    {
        params = (move_params *)calloc(1, sizeof(move_params));

        params->vpanel = vpanel;
        g_object_ref(G_OBJECT(vpanel));

        copy_vmpa(&params->addr, addr);

        g_idle_add_full(G_PRIORITY_DEFAULT_IDLE,
                        (GSourceFunc)do_move_in_main_loop,
                        params, (GDestroyNotify)delete_move_params);

    }

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move   = bascule d'affichage à manipuler.                    *
*                binary = binaire de rattachement chargé en mémoire.          *
*                                                                             *
*  Description : Applique un déplacement au sein d'une zone de code.          *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_db_move_apply(GDbMove *move, GLoadedBinary *binary)
{
    bool result;                            /* Bilan à faire remonter      */

    result = g_db_move_run(move, &move->dest);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move   = informations de déplacement à consulter.            *
*                binary = binaire chargé en mémoire à modifier.               *
*                                                                             *
*  Description : Annule un déplacement au sein d'une zone de code.            *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_db_move_cancel(GDbMove *move, GLoadedBinary *binary)
{
    bool result;                            /* Bilan à faire remonter      */

    result = g_db_move_run(move, &move->src);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move   = base d'éléments sur laquelle s'appuyer.             *
*                values = couples de champs et de valeurs à lier. [OUT]       *
*                count  = nombre de ces couples. [OUT]                        *
*                                                                             *
*  Description : Constitue les champs destinés à une insertion / modification.*
*                                                                             *
*  Retour      : Etat du besoin en sauvegarde.                                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_db_move_prepare_db_statement(const GDbMove *move, bound_value **values, size_t *count)
{
    bool status;                            /* Bilan d'opération initiale  */

    status = G_DB_ITEM_CLASS(g_db_move_parent_class)->prepare_stmt(G_DB_ITEM(move), values, count);
    if (!status) return false;

    status = prepare_vmpa_db_statement(&move->src, "src", values, count);
    if (!status) return false;

    status = prepare_vmpa_db_statement(&move->dest, "dest", values, count);
    if (!status) return false;

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : move   = bascule d'affichage à charger depuis les réponses.  *
*                values = tableau d'éléments à consulter.                     *
*                count  = nombre de descriptions renseignées.                 *
*                                                                             *
*  Description : Charge les valeurs utiles pour un déplacement dans du code.  *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_db_move_load(GDbMove *move, const bound_value *values, size_t count)
{
    bool result;                            /* Bilan à faire remonter      */

    result = G_DB_ITEM_CLASS(g_db_move_parent_class)->load(G_DB_ITEM(move), values, count);

    result &= load_vmpa(&move->src, "src", values, count);

    result &= load_vmpa(&move->dest, "dest", values, count);

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                        DEFINITION DE LA COLLECTION ASSOCIEE                        */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une collection de déplacements dans du code. */
G_DEFINE_TYPE(GMoveCollection, g_move_collection, G_TYPE_DB_COLLECTION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des collections de déplacements.        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_move_collection_class_init(GMoveCollectionClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GDbCollectionClass *collec;             /* Encore une autre vision...  */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_move_collection_dispose;
    object->finalize = (GObjectFinalizeFunc)g_move_collection_finalize;

    collec = G_DB_COLLECTION_CLASS(klass);

    collec->create_table = (collec_create_db_table_fc)g_move_collection_create_db_table;
    collec->setup_load = (collec_setup_load_fc)g_move_collection_setup_load;
    collec->has_key = (collec_has_key_fc)g_move_collection_has_key;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collec = instance à initialiser.                             *
*                                                                             *
*  Description : Initialise une collection de déplacements.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_move_collection_init(GMoveCollection *collec)
{
    G_DB_COLLECTION(collec)->featuring = DBF_MOVES;
    G_DB_COLLECTION(collec)->type = G_TYPE_DB_MOVE;
    G_DB_COLLECTION(collec)->name = "Moves";

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collec = instance d'objet GLib à traiter.                    *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_move_collection_dispose(GMoveCollection *collec)
{
    G_OBJECT_CLASS(g_move_collection_parent_class)->dispose(G_OBJECT(collec));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collec = instance d'objet GLib à traiter.                    *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_move_collection_finalize(GMoveCollection *collec)
{
    G_OBJECT_CLASS(g_move_collection_parent_class)->finalize(G_OBJECT(collec));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Crée une collection dédiée aux déplacements dans du code.    *
*                                                                             *
*  Retour      : Collection mise en place.                                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GMoveCollection *g_move_collection_new(void)
{
    GMoveCollection *result;            /* Instance à retourner        */

    result = g_object_new(G_TYPE_MOVE_COLLECTION, NULL);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collec = ensemble d'éléments spectateur des opérations.      *
*                db     = accès à la base de données.                         *
*                                                                             *
*  Description : Crée la table des basculements dans une base de données.     *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_move_collection_create_db_table(const GMoveCollection *collec, sqlite3 *db)
{
    char *sql;                              /* Requête à exécuter          */
    int ret;                                /* Bilan de la création        */
    char *msg;                              /* Message d'erreur            */

    sql = "CREATE TABLE Moves ("                \
             SQLITE_DB_ITEM_CREATE ", "         \
             SQLITE_VMPA_CREATE("src") ", "     \
             SQLITE_VMPA_CREATE("dest")         \
          ");";

    ret = sqlite3_exec(db, sql, NULL, NULL, &msg);
    if (ret != SQLITE_OK)
    {
        fprintf(stderr, "sqlite3_exec(): %s\n", msg);
        sqlite3_free(msg);
    }

    return (ret == SQLITE_OK);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collec = ensemble d'éléments à consulter.                    *
*                values = tableau d'éléments à compléter. [OUT]               *
*                count  = nombre de descriptions renseignées. [OUT]           *
*                                                                             *
*  Description : Décrit les colonnes utiles à un chargement de données.       *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_move_collection_setup_load(GMoveCollection *collec, bound_value **values, size_t *count)
{
    bool status;                            /* Bilan d'une préparation     */

    status = G_DB_COLLECTION_CLASS(g_move_collection_parent_class)->setup_load(G_DB_COLLECTION(collec), \
                                                                                   values, count);
    if (!status) return false;

    if (!setup_load_for_vmpa(NULL, "src", values, count))
        return false;

    if (!setup_load_for_vmpa(NULL, "dest", values, count))
        return false;

    return true;


}


/******************************************************************************
*                                                                             *
*  Paramètres  : collec = ensemble d'éléments à consulter.                    *
*                ap     = clef identifiant de manière unique un élément.      *
*                                                                             *
*  Description : Détermine si un élément est déjà présent ou non.             *
*                                                                             *
*  Retour      : Elément trouvé ou NULL si aucun.                             *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GDbItem *g_move_collection_has_key(GMoveCollection *collec, va_list ap)
{
    GDbItem *result;                        /* Bilan à retourner           */
    vmpa2t *ref;                            /* Adresse de référence        */
    GList *items;                           /* Eléments déjà en place      */
    GList *iter;                            /* Boucle de parcours          */
    GDbMove *bm;                        /* Signet à consulter          */

    result = NULL;

    ref = va_arg(ap, vmpa2t *);

    items = g_db_collection_list_items(G_DB_COLLECTION(collec));

    for (iter = g_list_first(items); iter != NULL && result == NULL; iter = g_list_next(iter))
    {
        bm = G_DB_MOVE(iter->data);

        if (cmp_vmpa(&bm->src, ref) != 0)

            /**
             * Un verrou est sensé être posé, donc il n'y a pas besoin
             * de toucher au décompte des références car l'élément trouvé
             * ne peut pas être supprimé.
             */
            result = G_DB_ITEM(bm);

    }

    return result;

}
